#      MCcoordstore - simple web based store for Minecraft points of interest
#      Copyright (C) 2021  Daniel Philip Weatherill

#      This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.

#      This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.

#      You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from flask_login import LoginManager
from flask import g
from .db import User
from flask import jsonify, current_app
from itsdangerous import BadSignature, SignatureExpired

lman = LoginManager()
@lman.user_loader
def load_user(user_id: str):
    return User.query.get(int(user_id))

@lman.request_loader
def request_loader(request):
    token = request.args.get("token")
    if not token:
        #try header based auth
        token = request.headers.get("Authorization")
        if token:
            token = token.replace("Basic ", "", 1)
    
    if token:
        try:
            usr = User.verify_api_token(current_app, token)
            return usr
        except BadSignature:
            g.login_error_str = "bad signature in JWT"
            return None
        except SignatureExpired:
            g.login_error_str = "JWT signature expired"
    else:
        g.login_error_str = "no API token provided"
        return None


lman.login_message = "you need to login to perform this action"
lman.login_message_category = "flash-error"
lman.login_view = "login"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#      MCcoordstore - simple web based store for Minecraft points of interest
#      Copyright (C) 2021  Daniel Philip Weatherill

#      This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.

#      This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.

#      You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



from flask.cli import with_appcontext
from flask_sqlalchemy import SQLAlchemy
from datetime import datetime, timezone
from sqlalchemy import Enum, MetaData
import click
from typing import Sequence
from werkzeug.security import generate_password_hash
from flask_login import UserMixin
from itsdangerous import TimedJSONWebSignatureSerializer
import enum
from random import randint
from flask_migrate import stamp
import os

convention = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
}



metadata = MetaData(naming_convention=convention)
db = SQLAlchemy(metadata=metadata)

USER_MIN_ALTERNATE_ID = 1000
USER_MAX_ALTERNATE_ID = 1000000


def _server_random_sqlite_range():
    return "abs(random() % ( {max} - {min} )) + {min}".format(min=USER_MIN_ALTERNATE_ID, max=USER_MAX_ALTERNATE_ID)


class User(UserMixin, db.Model):
    __tablename__ = "user"
    MIN_ALTERNATE_ID=USER_MIN_ALTERNATE_ID
    MAX_ALTERNATE_ID=USER_MAX_ALTERNATE_ID
    
    userid = db.Column(db.Integer, primary_key = True)
    username = db.Column(db.String(80), unique=True, nullable=False)
    displayname = db.Column(db.String(80), unique=True, nullable=False)
    hashed_pw = db.Column(db.String, nullable=False)
    is_admin = db.Column(db.Boolean, default=False)
    
    alternate_id = db.Column(db.Integer, unique=True, nullable=False, server_default=_server_random_sqlite_range())
    
    
    @classmethod
    def create_new_user(cls, username: str, displayname: str, password: str) -> "User":
        pwhash = generate_password_hash(password)
        altid = cls.random_unique_alternateid()
        return User(username = username, displayname=displayname, 
                    hashed_pw=pwhash, alternate_id=altid)

    def get_id(self):
        return str(self.userid)

    def generate_api_token(self, app, expiration_seconds: int = 86400) -> bytes:
        secret_key = app.config["SECRET_KEY"]
        ser = TimedJSONWebSignatureSerializer(secret_key, expiration_seconds)
        return ser.dumps({"userid" : self.alternate_id}).decode("ASCII")
    
    @classmethod
    def verify_api_token(cls, app, token: bytes):
        secret_key = app.config["SECRET_KEY"]
        ser = TimedJSONWebSignatureSerializer(secret_key)
        data = ser.loads(token)
        user = cls.query.filter_by(alternate_id=data["userid"]).one()
        return user        

    @classmethod
    def random_unique_alternateid(cls):
        randchoice = randint(cls.MIN_ALTERNATE_ID, cls.MAX_ALTERNATE_ID)
        rcquery = lambda s :  cls.query.filter_by(alternate_id=s).limit(1).first()
        while rcquery(randchoice) is not None:
            randchoice = randint(cls.MIN_ALTERNATE_ID, cls.MAX_ALTERNATE_ID)
        return randchoice





class RenderStyle(db.Model):
    __tablename__ = "renderstyle"
    styleid = db.Column(db.Integer, primary_key = True)
    name = db.Column(db.String(80), nullable=True)
    style = db.Column(db.PickleType)


class CoordType(enum.IntEnum):
    OVERWORLD = 1,
    NETHER = 2,
    END = 3,
    UNDEFINED = 0

POI_NAME_LOOKUP = {CoordType.OVERWORLD : "Overworld",
                   CoordType.NETHER : "The Nether",
                   CoordType.END : "The End",
                   CoordType.UNDEFINED : "undefined"}



poi_tag_association = db.Table("poi_tag_association", db.metadata,
                            db.Column("poi_id", db.ForeignKey("pointofinterest.poiid"), primary_key=True),
                            db.Column("tag_id", db.ForeignKey("tag.tagid"), primary_key=True))


class PointOfInterest(db.Model):
    __tablename__ = "pointofinterest"
    poiid = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(80), nullable=False)
    create_date = db.Column(db.DateTime(timezone=True), nullable=False, default=lambda : datetime.now(timezone.utc))
    public = db.Column(db.Boolean, default=False)

    userid = db.Column(db.Integer, db.ForeignKey("user.userid"), nullable=False)
    user = db.relationship("User", backref=db.backref("pois"))
    style = db.Column(db.Integer, db.ForeignKey("renderstyle.styleid"))

    coord_x = db.Column(db.Integer)
    coord_y = db.Column(db.Integer)
    coord_z = db.Column(db.Integer)
    
    coordtype = db.Column(Enum(CoordType, default=str(CoordType.UNDEFINED.value), nullable=False,
                               server_default=str(CoordType.OVERWORLD.value),
                               values_callable = lambda enum: [str(_.value) for _ in enum]))

    @property
    def coords(self):
        return (self.coord_x, self.coord_y, self.coord_z)
    
    @property
    def typename(self):
        return POI_NAME_LOOKUP[self.coordtype]
    
    @coords.setter
    def coords(self, val: Sequence[int]):
        if len(val) != 3:
            raise ValueError("invalid length of coordinate")
        self.coord_x = val[0]
        self.coord_y = val[1]
        self.coord_z = val[2]
        
    def serialize_to_dict(self):
        return {"id" : self.poiid,
                "name" : self.name,
                "created" : str(self.create_date),
                "public" : self.public,
                "username" : self.user.username,
                "coord" : self.coords,
                "coord_type" : POI_NAME_LOOKUP[self.coordtype]
                }

class Tag(db.Model):
    __tablename__ = "tag"
    tagid = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, nullable=False)
    create_date = db.Column(db.DateTime(timezone=True), nullable=False, default= lambda: datetime.now(timezone.utc))
    
    userid = db.Column(db.Integer, db.ForeignKey("user.userid"), nullable=False)
    owning_user = db.relationship("User", backref=db.backref("tags"))
    style = db.Column(db.Integer, db.ForeignKey("renderstyle.styleid"))
    
    pois = db.relationship("PointOfInterest", secondary=poi_tag_association, backref="tags")


def get_db(app):
    db.init_app(app)
    return db


def create_db(admin_pass, db):
    db_path = db.engine.url
    print(f"creating new database with path: {db_path}")
    db.create_all()

    print("creating default admin and guest users")
    admin_user = User.create_new_user("admin", "admin", admin_pass)
    admin_user.is_admin = True
    guest_user = User.create_new_user("guest", "guest", "guest")

    db.session.add(admin_user)
    db.session.add(guest_user)
    db.session.commit()
    


@click.command("create-db")
@click.argument("admin_pass")
@with_appcontext
def create_db_command(admin_pass):
    from flask import current_app
    db.init_app(current_app)
    create_db(admin_pass, db)
    migdir = os.path.abspath(os.path.dirname(__file__)) + os.path.sep + "migrations"
    stamp(directory=migdir)
    


@click.command("change-pw")
@click.argument("username")
@click.argument("new_pass")
@with_appcontext
def change_pw_command(username, new_pass):
    from flask import current_app
    db.init_app(current_app)

    usr = User.query.filter_by(username=username).first()
    pwhash = generate_password_hash(new_pass)
    usr.hashed_pw = pwhash
    db.session.commit()

    




#      MCcoordstore - simple web based store for Minecraft points of interest
#      Copyright (C) 2021  Daniel Philip Weatherill

#      This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.

#      This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.

#      You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from flask import Flask, g
import os
from typing import Dict, Any
from .db import create_db_command, User, change_pw_command, get_db, create_db
from flask import current_app
from flask_login import LoginManager
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy
import base64
from .api_poi import poi_api, setup_api
from .login import lman
from .db import User, RenderStyle, CoordType, PointOfInterest, Tag



CONFIG_ENVVAR = "MCCOORDSTORE_CONFIG"


def create_app(test_config: Dict[str, Any]=None):
    app = Flask(__name__, instance_relative_config=True)
    
    if test_config is None:
        if os.environ.get(CONFIG_ENVVAR) is None:
            print("WARNING: using default test config! DO NOT USE IN PRODUCTION!")
            app.config.from_object("MCcoordstore.default_config.DefaultConfig")
        else:
            app.config.from_envvar(CONFIG_ENVVAR, silent=False)
    else:
        app.config.from_mapping(test_config)

    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass
    
    app.cli.add_command(create_db_command)
    app.cli.add_command(change_pw_command)

    
    print("registering poi_api blueprint")
    
    app.register_blueprint(poi_api, url_prefix="/api")
    
    db = get_db(app)
    lman.init_app(app)
    migrate = Migrate(app,db, render_as_batch=True)

    with app.app_context():
        setup_api(app, db)
    return app

    



#      MCcoordstore - simple web based store for Minecraft points of interest
#      Copyright (C) 2021  Daniel Philip Weatherill

#      This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.

#      This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.

#      You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from flask import Blueprint, jsonify, current_app, request
from flask.blueprints import BlueprintSetupState
from flask_login import LoginManager, login_required, current_user
from .db import User, PointOfInterest
import base64
from MCcoordstore import db
from flask_restless import APIManager, ProcessingException
from .serializers import POISerializer

poi_api = Blueprint("poi", __name__)


@poi_api.route("/gettoken")
@login_required
def get_token():
    if current_user is None:
        return jsonify({"error" : "Require logged in user to get API token"})
    
    token = current_user.generate_api_token(current_app)
    
    return jsonify({"token" : token,
                    "error" : ""})
    

def api_auth_check(*args, **kwargs):
    if not current_user.is_authenticated:
        raise ProcessingException(detail="not authenticated", status=401)
        
def remove_nonpublic_collection_poi(filters, **kwargs):
    if not current_user.is_authenticated:
        filters.append({"name": "public", "op": "eq", "val" : True})

    
def setup_api(app, db):
    manager = APIManager(app, session=db.session)
    
    
    poi_preproc = {"GET_COLLECTION" : [remove_nonpublic_collection_poi],
                   "GET_RESOURCE" : [api_auth_check] }
    manager.create_api(PointOfInterest, methods=["GET"], collection_name="poi",
                       additional_attributes=["coords", "typename"], exclude=["coord_x","coord_y","coord_z", "tags", "poiid"],
                       preprocessors=poi_preproc)
    
    user_preproc = {"GET_COLLECTION" : [api_auth_check],
                    "GET_RESOURCE" : [api_auth_check]}
    
    manager.create_api(User, methods=["GET"],
                       exclude=["hashed_pw", "tags", "userid", "username"],
                       preprocessors=user_preproc)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec 22 16:51:02 2021

@author: danw
"""

from flask_restless import Serializer
from flask_login import current_user
from .db import PointOfInterest, User
from typing import Dict
from sqlalchemy import inspect
from abc import abstractmethod

class APISerializer(Serializer):
    DB_MODEL_CLASS = None
    def __init__(self):
        if self.DB_MODEL_CLASS is None:
            raise TypeError("no DB_MODEL_CLASS defined")
        self._inspector = inspect(self.DB_MODEL_CLASS)
        
    @abstractmethod
    def _init_override(self):
        pass

class POISerializer(APISerializer):
    
        
        
    @property
    def attributes_columns(self):
        return set(self.attributes)
    
    @property
    def relationships_columns(self):
        return set(self.relationships)
    
    def serialize(instance: PointOfInterest, only=None) -> Dict:
        pass

